unit BudgAcct;

interface

uses
  SysUtils, Classes, ObjColl;

type
  TAcct = class( TComponent )
  private
    FName           : string;
    FAmtPerPaycheck : double;
    FBalance        : double;
  private
    function GetAmtPerMonth : double;
    procedure SCGetString( var msg : TStringMessage ); message SC_GETSTRING;
    procedure SCSetString( var msg : TStringMessage ); message SC_SETSTRING;
  public
    procedure Deposit( amount : double );
    procedure Distribute( var amount : double );
    procedure Withdrawl( amount : double );
  published
    property Name : string read FName write FName;
    property AmtPerPaycheck : double read FAmtPerPaycheck write FAmtPerPaycheck;
    property AmtPerMonth : double read GetAmtPerMonth;
    property Balance : double read FBalance write FBalance;
  end;

  TAccounts = class( TComponent )
  private
    FAccounts    : TObjectCollection;
    FPaycheckAmt : double;
    FCash        : double;
    FCreditCard  : double;
  private
    function GetAccount( index : integer ) : TAcct;
    function GetAccountByName( const name : string ) : TAcct;
    function GetBalance : double;
    function GetBudgetMonthly : double;
    function GetBudgetPaycheck : double;
    function GetCount : integer;
    procedure ReadAccounts( reader : TReader );
    procedure WriteAccounts( writer : TWriter );
  protected
    procedure DefineProperties( filer: TFiler ); override;
  public
    constructor Create( aOwner : TComponent ); override;
    destructor Destroy; override;
    procedure AddAccount( const name : string; balance, amtPerPaycheck : double );
    procedure ClearBalances;
    procedure DeleteAccount( index : integer );
    procedure DistributePaycheck;
    procedure InsertAccount( index : integer; const name : string;
      balance, amtPerPaycheck : double );
  public
    property Account[index:integer] : TAcct read GetAccount;
    property AccountByName[const name:string] : TAcct read GetAccountByName;
  published
    property Balance : double read GetBalance;
    property BudgetMonthly : double read GetBudgetMonthly;
    property BudgetPaycheck : double read GetBudgetPaycheck;
    property Count : integer read GetCount;
    property PaycheckAmt : double read FPaycheckAmt write FPaycheckAmt;
    property Cash : double read FCash write FCash;
    property CreditCard : double read FCreditCard write FCreditCard;
  end;


implementation

{-TAcct------------------------------------------------------------------------}
procedure TAcct.Deposit( amount : double );
begin
  FBalance := FBalance + amount;
end;

procedure TAcct.Distribute( var amount : double );
var
  taken : double;
begin
  if (amount <= 0.0) then exit;
  taken := FAmtPerPaycheck;
  if (taken > amount) then taken := amount;
  Deposit( taken );
  amount := amount - taken;
end;

function TAcct.GetAmtPerMonth : double;
begin
  Result := Round( (FAmtPerPaycheck * 26) / 12 );
end;

procedure TAcct.SCGetString( var msg : TStringMessage );
begin
  msg.s := FName;
end;

procedure TAcct.SCSetString( var msg : TStringMessage );
begin
  FName := msg.s;
end;

procedure TAcct.Withdrawl( amount : double );
begin
  FBalance := FBalance - amount;
end;


{-TAccounts--------------------------------------------------------------------}
constructor TAccounts.Create( aOwner : TComponent );
begin
  inherited Create( aOwner );
  FAccounts := TObjectCollection.Create( nil );
end;

destructor TAccounts.Destroy;
begin
  FAccounts.Free;
  inherited Destroy;
end;

procedure TAccounts.AddAccount( const name : string; balance, amtPerPaycheck : double );
var
  item : TAcct;
begin
  item := TAcct.Create( nil );
  item.Name := name;
  item.AmtPerPaycheck := amtPerPaycheck;
  item.Balance := balance;
  FAccounts.AddObject( '', item );
end;

procedure TAccounts.ClearBalances;
  function ClearBalance( index : integer; item : TAcct ) : integer; far;
  begin
    item.Balance := 0.0;
    Result := 0;
  end;
begin
  FAccounts.ForAll( @ClearBalance );
end;

procedure TAccounts.DefineProperties( filer: TFiler );
begin
  filer.DefineProperty( 'Accounts', ReadAccounts, WriteAccounts, true );
end;

procedure TAccounts.DeleteAccount( index : integer );
begin
  if (index < FAccounts.Count) then
    FAccounts.Delete( index );
end;

procedure TAccounts.DistributePaycheck;
var
  amt : double;
  function DistributeAmt( index : integer; item : TAcct ) : integer; far;
  begin
    item.Distribute( amt );
    Result := 0;
  end;
begin
  amt := FPaycheckAmt;
  { distribute amount to accounts }
  FAccounts.ForAll( @DistributeAmt );
  { put remainder in savings }
  GetAccountByName( 'Savings' ).Deposit( amt );
end;

function TAccounts.GetAccount( index : integer ) : TAcct;
begin
  Result := TAcct(FAccounts.Objects[index]);
end;

function TAccounts.GetAccountByName( const name : string ) : TAcct;
begin
  Result := TAcct(FAccounts.Objects[FAccounts.IndexOf(name)]);
end;

function TAccounts.GetBalance : double;
var
  amt : double;
  function AddBalance( index : integer; item : TAcct ) : integer; far;
  begin
    amt := amt + item.Balance;
    Result := 0;
  end;
begin
  amt := 0.0;
  FAccounts.ForAll( @AddBalance );
  Result := amt;
end;

function TAccounts.GetBudgetMonthly : double;
var
  amt : double;
  function AddAmt( index : integer; item : TAcct ) : integer; far;
  begin
    amt := amt + item.AmtPerMonth;
    Result := 0;
  end;
begin
  amt := 0.0;
  FAccounts.ForAll( @AddAmt );
  Result := amt;
end;

function TAccounts.GetBudgetPaycheck : double;
var
  amt : double;
  function AddAmt( index : integer; item : TAcct ) : integer; far;
  begin
    amt := amt + item.AmtPerPaycheck;
    Result := 0;
  end;
begin
  amt := 0.0;
  FAccounts.ForAll( @AddAmt );
  Result := amt;
end;

function TAccounts.GetCount : integer;
begin
  Result := FAccounts.Count;
end;

procedure TAccounts.InsertAccount( index : integer; const name : string;
  balance, amtPerPaycheck : double );
var
  item : TAcct;
begin
  item := TAcct.Create( nil );
  item.Name := name;
  item.AmtPerPaycheck := amtPerPaycheck;
  item.Balance := balance;
  FAccounts.InsertObject( index, '', item );
end;

procedure TAccounts.ReadAccounts( reader : TReader );
begin
  reader.ReadListBegin;
  while not reader.EndOfList do
    with reader do
      AddAccount( ReadString, ReadFloat, ReadFloat );
  reader.ReadListEnd;
end;

procedure TAccounts.WriteAccounts( writer : TWriter );
  function WriteIt( index : integer; item : TAcct ) : integer; far;
  begin
    with item do
    begin
      writer.WriteString( Name );
      writer.WriteFloat( Balance );
      writer.WriteFloat( AmtPerPaycheck );
    end;
    Result := 0;
  end;
begin
  writer.WriteListBegin;
  FAccounts.ForAll( @WriteIt );
  writer.WriteListEnd;
end;

end.
