unit ObjColl;

(* 
  portions Copyright Borland International
    (ForAll & QuickSort)
*)

interface

uses
  Classes;

const
  SC_GETSTRING = $7FFF;
  SC_SETSTRING = $7FFE;

type
  TGetStringEvent = procedure( Item : TObject; var s : string ) of object;
  TSetStringEvent = procedure( Item : TObject; const s : string ) of object;

  TStringMessage = record
    msg : word;
    s   : string;
  end;

  TObjectCollection = class(TStrings)
  private
    FClass       : TClass;
    FList        : TList;
    FOnGetString : TGetStringEvent;
    FOnSetString : TSetStringEvent;
    FOnChange    : TNotifyEvent;
    FOnChanging  : TNotifyEvent;
    procedure QuickSort(L, R: Integer);
  protected
    procedure Changed; virtual;
    procedure Changing; virtual;
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure Put(Index: Integer; const S: string); override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetUpdateState(Updating: Boolean); override;
  public
    constructor Create( aClass : TClass );
    destructor Destroy; override;
    function Add(const S: string): Integer; override;
    function AddObject(const S: string; AObject: TObject): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Exchange(Index1, Index2: Integer); override;
    function ForAll( ApplyFunction : pointer ) : integer;
    procedure Insert(Index: Integer; const S: string); override;
    {procedure InsertObject(Index: Integer; const S: string; AObject: TObject); override;}
    procedure Sort; virtual;
    property OnGetString: TGetStringEvent read FOnGetString write FOnGetString;
    property OnSetString: TSetStringEvent read FOnSetString write FOnSetString;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnChanging: TNotifyEvent read FOnChanging write FOnChanging;
  end;

  {
    - if the objects that TObjectColletion owns respond to SC_GETSTRING then
      then returned string is used in the Get method
    - if the objects that TObjectColletion owns respond to SC_SETSTRING then
      then string is passed to the object in the Put method
    - if an class is passed to Create then that class is used to create objects
      when new strings are added or inserted, if you don't want TObjectCollecion
      create objects then pass nil to Create
    - existing classes that do not respond to SC_GET/SETSTRING can be used
      in TObjectCollection as well, if you want a string representation of these
      objects then set OnGetString and/or OnSetString events
    - TObjectCollection is just a container owning all of its objects.  If you
      do not need the string functions of TStrings or you do not need to use
      TObjectCollection in correlation with other TStrings types then it works
      just fine ignoring SC_GET/SETSTRING and OnGet/SetString.  Sometimes it is
      nice to be able to represent an objects data as a string for the purpose
      of assigning it to a TListObj, TComboBox, etc.
    - The ForAll calls ApplyFuntion for each non-nil object in the collection:
        ApplyFunction( index : integer, item : 'pointer' ) : integer;
      returns non-zero to halt loop
      this is similar to (but not the same as) TCollection.ForEach in BP7
    - objects that the collection owns are freed when Clear, Delete, Destroy,
      or Put are called (similar to TCollection)
  }

implementation

uses
  SysUtils;

constructor TObjectCollection.Create( aClass : TClass );
begin
  FClass := aClass;
  FList := TList.Create;
end;

destructor TObjectCollection.Destroy;
begin
  FOnGetString := nil;
  FOnSetString := nil;
  FOnChange := nil;
  FOnChanging := nil;
  if FList <> nil then
  begin
    Clear;
    FList.Free;
  end;
end;

function TObjectCollection.Add(const S: string): Integer;
var
  p : TObject;
begin
  Result := FList.Count;
  Changing;
  if (FClass <> nil) then
    p := FClass.Create
  else
    p := nil;
  FList.Expand.Insert( Result, p );
  if (p <> nil) and (S <> '') then
    Put( Result, S );
  Changed;
end;

function TObjectCollection.AddObject(const S: string; AObject: TObject): Integer;
begin
  Result := FList.Count;
  Changing;
  if (AObject = nil) then
    if (FClass <> nil) then
      AObject := FClass.Create;
  FList.Expand.Insert( Result, AObject );
  if (AObject <> nil) and (S <> '') then
    Put( Result, S );
  Changed;
end;

procedure TObjectCollection.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TObjectCollection.Changing;
begin
  if Assigned(FOnChanging) then FOnChanging(Self);
end;

procedure TObjectCollection.Clear;
var
  I: Integer;
begin
  Changing;
  for I := 0 to FList.Count - 1 do
    TObject(FList[I]).Free;
  FList.Clear;
  Changed;
end;

procedure TObjectCollection.Delete(Index: Integer);
begin
  Changing;
  TObject(FList[Index]).Free;
  FList.Delete(Index);
  Changed;
end;

procedure TObjectCollection.Exchange(Index1, Index2: Integer);
begin
  Changing;
  FList.Exchange(Index1, Index2);
  Changed;
end;

function TObjectCollection.ForAll( ApplyFunction : pointer ) : integer;
{ Taken from the Grids unit }
var
  callerBP : cardinal;
  cntr     : integer;
  p        : pointer;
begin
  Result := 0;
  asm
    mov   ax,[bp]                       { Set up stack frame for local }
    and   al,$FE
    mov   callerBP,ax
  end;
  for cntr := 0 to FList.Count - 1 do
  begin
    p := FList.Items[cntr];
    if (p <> nil) then
    begin
      { ret := ApplyFunction(index, itemptr); }
      asm
        push  cntr
        push  word ptr p+2       { .Seg }
        push  word ptr p         { .Ofs }
        push  callerBP           { Outer proc's BP for nested proc }
        call  ApplyFunction
        mov   @Result,ax
      end;
      if (Result <> 0) then break;
    end;
  end;
end;

function TObjectCollection.Get(Index: Integer): string;
var
  msg : TStringMessage;
  p   : TObject;
begin
  p := TObject(FList[Index]);
  if (p <> nil) then
  begin
    msg.msg := SC_GETSTRING;
    msg.s := '';
    p.Dispatch( msg );
    Result := msg.s;
  end
  else
    Result := '';
  if Assigned(FOnGetString) then
    FOnGetString( p, Result );
end;

function TObjectCollection.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TObjectCollection.GetObject(Index: Integer): TObject;
begin
  Result := TObject(FList[Index]);
end;

procedure TObjectCollection.Insert(Index: Integer; const S: string);
var
  p : TObject;
begin
  Changing;
  if (FClass <> nil) then
    p := FClass.Create
  else
    p := nil;
  FList.Expand.Insert( Index, p );
  if (p <> nil) and (S <> '') then
    Put( Index, S );
  Changed;
end;

{
--
we cannot override this procedure because it is not virtual in TStrings (but
AddObject is) !!
--
procedure TObjectCollection.InsertObject(Index: Integer; const S: string;
  AObject : TObject);
begin
  Changing;
  if (AObject = nil) then
    if (FClass <> nil) then
      AObject := FClass.Create;
  FList.Expand.Insert( Index, AObject );
  if (AObject <> nil) and (S <> '') then
    Put( Index, S );
  Changed;
end;
}

procedure TObjectCollection.Put(Index: Integer; const S: string);
var
  msg : TStringMessage;
  p   : TObject;
begin
  Changing;
  p := TObject(FList[Index]);
  if (p <> nil) then
  begin
    msg.msg := SC_SETSTRING;
    msg.s := S;
    p.Dispatch( msg );
  end
  else
    msg.s := S;
  if Assigned(FOnSetString) then
    FOnSetString( p, msg.s );
  Changed;
end;

procedure TObjectCollection.PutObject(Index: Integer; AObject: TObject);
begin
  Changing;
  if (Index >= FList.Count) then
    FList.Expand.Insert( FList.Count, AObject )
  else
  begin
    TObject(FList[Index]).Free;
    FList[Index] := AObject;
  end;
  Changed;
end;

procedure TObjectCollection.QuickSort(L, R: Integer);
{ Taken from the TStringsList }
var
  I, J: Integer;
  S: string;
begin
  I := L;
  J := R;
  S := Get( (L + R) shr 1 );
  repeat
    while AnsiCompareText(Get(I), S) < 0 do Inc(I);
    while AnsiCompareText(Get(J), S) > 0 do Dec(J);
    if I <= J then
    begin
      FList.Exchange(I, J);
      Inc(I);
      Dec(J);
    end;
  until I > J;
  if L < J then QuickSort(L, J);
  if I < R then QuickSort(I, R);
end;

procedure TObjectCollection.SetUpdateState(Updating: Boolean);
begin
  if Updating then Changing else Changed;
end;

procedure TObjectCollection.Sort;
begin
  if (FList.Count > 1) then
  begin
    Changing;
    QuickSort(0, FList.Count - 1);
    Changed;
  end;
end;


end.
