unit RptCol;

interface

uses
  SysUtils, WinTypes, WinProcs, Classes, Graphics, Printers;

type
  TJustify = (jfLeft, jfCenter, jfRight);
  TTitleString = string[80];
  TFormatString = string[50];

  TReportField = class
  private
    FTitle      : TTitleString;
    FFormat     : TFormatString;
    FWidth      : integer;
    FJustify    : TJustify;
    FFont       : TFont;
    FTitleFont  : TFont;
    FXWidth     : integer;
    FYHeight    : integer;
    FValue      : string;
    function GetTitleFont : TFont;
    procedure SetFont( aFont : TFont );
    procedure SetTitleFont( aFont : TFont );
    procedure SetWidth( aWidth : integer );
  protected
    curFont     : TFont;
  public
    constructor Create( const aTitle, aFormat : string; aFont : TFont;
                        aWidth : integer; aJustify : TJustify );
    destructor Destroy; override;
    procedure SetData( data : array of const );
    procedure SetTitle;
    function OutputText( canvas : TCanvas; pos : TPoint;
                         var height : integer ) : integer;
  public
    property FieldHeight : integer read FYHeight write FYHeight;
    property FieldWidth : integer read FXWidth write FXWidth;
    property Font : TFont read FFont write SetFont;
    property Format : TFormatString read FFormat write FFormat;
    property Text : string read FValue write FValue;
    property Title : TTitleString read FTitle write FTitle;
    property TitleFont : TFont read FTitleFont write SetTitleFont;
    property Width : integer read FWidth write SetWidth;
  end;


  TColumnReport = class
  private
    FFields      : TStringList;
    FPrinter     : TPrinter;
    FFont        : TFont;
    FPos         : TPoint;
    FLine        : integer;
    FLineHeight  : integer;
    FLineIndent  : double;
    FColumnSep   : double;
    FIsNewPage   : boolean;
    FLastField   : TReportField;
    function GetCanvas : TCanvas;
    function GetPage : integer;
    function GetField( const name : string ) : TReportField;
    procedure SetFieldFont( aFont : TFont );
    procedure SetFont( aFont : TFont );
    procedure SetTitleFont( aFont : TFont );
  protected
    fieldsToUse  : set of byte;
    pixelsInch   : integer;
    procedure PrintFields; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddField( const name, format : string;
                        width : integer; justify : TJustify ); virtual;
    procedure ClearFields;
    procedure SetFieldsToUse( const fieldNames : array of string );
    procedure Start; virtual;
    procedure Stop; virtual;
    procedure IncLine( numOfLines : integer );
    procedure IncPage;
    procedure NeedLines( numOfLines : integer );
    procedure PrintFooter; virtual;
    procedure PrintHeader; virtual;
    procedure PrintLines; virtual;
    procedure PrintLine( data : array of const ); virtual;
    procedure PrintString( const data : string ); virtual;
  public
    property Canvas : TCanvas read GetCanvas;
    property Fields[const name:string] : TReportField read GetField; default;
    property LastField : TReportField read FLastField;
    property IsNewPage : boolean read FIsNewPage;
    property Line : integer read FLine;
    property LineHeight : integer read FLineHeight;
    property Printer : TPrinter read FPrinter;

    property ColumnSeparation : double read FColumnSep write FColumnSep;
    property Font : TFont read FFont write SetFont;
    property TitleFont : TFont write SetTitleFont;
    property FieldFont : TFont write SetFieldFont;
    property LineIndent : double read FLineIndent write FLineIndent;
    property Page : integer read GetPage;
  end;




implementation


{------------------------------------------------------------------------------}
constructor TReportField.Create( const aTitle, aFormat : string; aFont : TFont;
                                 aWidth : integer; aJustify : TJustify );
begin
  inherited Create;
  FFont := TFont.Create;
  FFont.Assign( aFont );
  FTitleFont := FFont;
  FTitle := aTitle;
  FFormat := aFormat;
  FJustify := aJustify;
  FValue := '';
  curFont := FFont;
  SetWidth( aWidth );
end;

destructor TReportField.Destroy;
begin
  try
    if (FTitleFont <> FFont) then
      FTitleFont.Free;
    FFont.Free;
  finally
    inherited Destroy;
  end;
end;

function TReportField.GetTitleFont : TFont;
begin
  { FTitleFont is the same as FFont unless it is accessed }
  if (FTitleFont = FFont) then
  begin
    FTitleFont := TFont.Create;
    FTitleFont.Assign( FFont );
  end;
  Result := FTitleFont;
end;

function TReportField.OutputText( canvas : TCanvas; pos : TPoint;
                                  var height : integer ) : integer;
var
  textMetrics : TTextMetric;
  rect        : TRect;
  fmt         : integer;
begin
  canvas.Font := CurFont;
  try
    GetTextMetrics( canvas.Handle, textMetrics );
    if (FXWidth = 0) then
      FXWidth := ((textMetrics.tmMaxCharWidth +
        textMetrics.tmAveCharWidth * 2) div 3) * FWidth;
    FYHeight := textMetrics.tmHeight;
    height := FYHeight;
    fmt := DT_BOTTOM or DT_NOPREFIX or DT_SINGLELINE;
    if (FJustify = jfCenter) then
      fmt := fmt or DT_CENTER
    else if (FJustify = jfRight) then
      fmt := fmt or DT_RIGHT;
    DrawText( canvas.Handle, @FValue[1], -1,
              Bounds( pos.x,pos.y,FXWidth,height ), fmt );
  finally
    Result := FXWidth;
  end;
end;

procedure TReportField.SetData( data : array of const );
begin
  if (TVarRec(data[0]).VType = vtPointer) and
     (TVarRec(data[0]).VPointer = nil) then
    FValue := ''
  else
    if (FFormat[1] <> '%') and (TVarRec(data[0]).VType = vtExtended) then
    begin
      if (Uppercase(Copy(FFormat,1,5)) = 'DATE ') then
        FValue := FormatDateTime( Copy(FFormat,6,255), TVarRec(data[0]).VExtended^ )
      else
        if (Uppercase(Copy(FFormat,1,6)) = 'FLOAT ') then
          FValue := FormatFloat( Copy(FFormat,7,255), TVarRec(data[0]).VExtended^ )
    end
    else
      try
        FmtStr( FValue, FFormat, data );
      except
        on EConvertError do FValue := '*error*';
      end;
  FValue := FValue + #0;
  CurFont := FFont;
end;

procedure TReportField.SetFont( aFont : TFont );
begin
  FFont.Assign( aFont );
end;

procedure TReportField.SetTitleFont( aFont : TFont );
begin
  { FTitleFont is the same as FFont unless it is accessed }
  if (FTitleFont = FFont) then
    FTitleFont := TFont.Create;
  FTitleFont.Assign( aFont );
end;

procedure TReportField.SetTitle;
begin
  FValue := FTitle + #0;
  CurFont := FTitleFont;
end;

procedure TReportField.SetWidth( aWidth : integer );
begin
  FWidth := aWidth;
  FXWidth := 0;
  FYHeight := 0;
end;


{------------------------------------------------------------------------------}
constructor TColumnReport.Create;
begin
  inherited Create;
  FFields := TStringList.Create;
  FPrinter := TPrinter.Create;
  FPrinter.PrinterIndex := Printers.Printer.PrinterIndex;
  FFont := TFont.Create;
  FFont.Assign( FPrinter.Canvas.Font );
  FLine := 0;
  fieldsToUse := [0..255];
  pixelsInch := Font.PixelsPerInch;
  FPos.x := 0;
  FPos.y := pixelsInch div 4;
  FLineHeight := pixelsInch div 6;
  FLineIndent := 0.25;
  FColumnSep := 0.10;
  FIsNewPage := true;
  FLastField := nil;
end;

destructor TColumnReport.Destroy;
begin
  Stop;
  ClearFields;
  try
    FFields.Free;
    FPrinter.Free;
    FFont.Free;
  finally
    inherited Destroy;
  end;
end;

procedure TColumnReport.AddField( const name, format : string;
                                  width : integer; justify : TJustify );
begin
  FLastField := TReportField.Create( name, format, FFont, width, justify );
  FFields.AddObject( name, FLastField );
end;

procedure TColumnReport.ClearFields;
var
  cntr : integer;
begin
  if (FFields.Count > 0) then
    for cntr := 0 to FFields.Count - 1 do
      FFields.Objects[cntr].Free;
  FFields.Clear;
  FLastField := nil;
end;

function TColumnReport.GetCanvas : TCanvas;
begin
  Result := FPrinter.Canvas;
end;

function TColumnReport.GetField( const name : string ) : TReportField;
begin
  with FFields do Result := Objects[IndexOf(name)] as TReportField;
end;

function TColumnReport.GetPage : integer;
begin
  Result := FPrinter.PageNumber;
end;

procedure TColumnReport.IncLine( numOfLines : integer );
begin
  FPos.x := 0;
  FPos.y := FPos.y + (LineHeight * numOfLines);
  FLine := FLine + numOfLines;
  NeedLines( 0 );
end;

procedure TColumnReport.IncPage;
begin
  PrintFooter;
end;

procedure TColumnReport.NeedLines( numOfLines : integer );
begin
  if FIsNewPage then exit;
  { see if we are at top or bottom of page }
  if (FPos.y + (LineHeight * numOfLines) + Round(pixelsInch*0.50) >
      FPrinter.PageHeight) then PrintFooter;
  if FIsNewPage then PrintHeader;
end;

procedure TColumnReport.PrintFields;
var
   cntr    : integer;
   height,
   hgt     : integer;
begin
  FPos.x := Round( FLineIndent * pixelsInch );
  height := 0;
  for cntr := 0 to FFields.Count - 1 do
    if (cntr in fieldsToUse) then
      with FFields.Objects[cntr] as TReportField do
      begin
        FPos.x :=
          FPos.x + OutputText( FPrinter.Canvas, FPos,hgt ) +
          Round( FColumnSep * pixelsInch );
        if (hgt > height) then height := hgt;
      end;
  FLineHeight := height + 2;
  IncLine( 1 );
end;

procedure TColumnReport.PrintFooter;
begin
  FPrinter.NewPage;
  FPos.x := 0;
  FPos.y := pixelsInch div 4;
  FLine := 0;
  FIsNewPage := true;
end;

procedure TColumnReport.PrintHeader;
var
  cntr : integer;
begin
  NeedLines( 3 );
  { print column headers }
  for cntr := 0 to FFields.Count - 1 do
    if (cntr in fieldsToUse) then
       (FFields.Objects[cntr] as TReportField).SetTitle;
  PrintFields;
  PrintLines;
  FIsNewPage := false;
end;

procedure TColumnReport.PrintLine( data : array of const );
var
  cntr     : integer;
  dataCntr : integer;
begin
  if FIsNewPage then PrintHeader;
  NeedLines( 1 );
  { set field data }
  dataCntr := 0;
  for cntr := 0 to FFields.Count - 1 do
    if (cntr in fieldsToUse) then
    begin
      (FFields.Objects[cntr] as TReportField).SetData( data[dataCntr] );
      Inc( dataCntr );
    end;
  { print the line }
  PrintFields;
end;

procedure TColumnReport.PrintLines;
var
  cntr : integer;
begin
  NeedLines( 1 );
  { print lines }
  FPos.x := Round( FLineIndent * pixelsInch );
  FPos.y := FPos.y + (FLineHeight div 2);
  for cntr := 0 to FFields.Count - 1 do
    if (cntr in fieldsToUse) then
    begin
      FPrinter.Canvas.MoveTo( FPos.x, FPos.y );
      FPos.x := FPos.x + (FFields.Objects[cntr] as TReportField).FieldWidth;
      FPrinter.Canvas.LineTo( FPos.x, FPos.y );
      FPos.x := FPos.x + Round( FColumnSep * pixelsInch );
    end;
  FPos.y := FPos.y - (FLineHeight div 2);
  IncLine( 1 );
end;

procedure TColumnReport.PrintString( const data : string );
var
  textMetrics : TTextMetric;
begin
  NeedLines( 1 );
  FPos.x := Round( FLineIndent * pixelsInch );
  FPrinter.Canvas.Font := FFont;
  GetTextMetrics( FPrinter.Canvas.Handle, textMetrics );
  FLineHeight := textMetrics.tmHeight + 2;
  FPrinter.Canvas.TextOut( FPos.x, FPos.y, data );
  IncLine( 1 );
end;

procedure TColumnReport.SetFieldsToUse( const fieldNames : array of string );
var
  cntr : integer;
  val  : integer;
begin
  fieldsToUse := [];
  for cntr := Low(fieldNames) to High(fieldNames) do
  begin
    val := FFields.IndexOf( fieldNames[cntr] );
    if (val >= 0) then
      Include( fieldsToUse, val );
  end;
end;

procedure TColumnReport.SetFieldFont( aFont : TFont );
var
  cntr : integer;
begin
  if (FFields.Count > 0) then
    for cntr := 0 to FFields.Count - 1 do
      TReportField(FFields.Objects[cntr]).Font := aFont;
end;

procedure TColumnReport.SetFont( aFont : TFont );
begin
  FFont.Assign( aFont );
end;

procedure TColumnReport.SetTitleFont( aFont : TFont );
var
  cntr : integer;
begin
  if (FFields.Count > 0) then
    for cntr := 0 to FFields.Count - 1 do
      TReportField(FFields.Objects[cntr]).TitleFont := aFont;
end;

procedure TColumnReport.Start;
begin
  FPrinter.BeginDoc;
end;

procedure TColumnReport.Stop;
begin
  {if (FLine > 0) then PrintFooter; -- causes blank page!}
  try
    if FPrinter.Printing then FPrinter.EndDoc;
  except
    on EPrinter do ;
  end;
end;

end.
